@@ -20,6 +20,8 @@
 
  *****************************************************************************)
 
+  (** Advanced scheduler and monad for server-oriented programming. *)
+
   (**  
     * {R {i {v 
     *        The bars could not hold me;
@@ -56,6 +58,13 @@
   * ['a] is the type of objects used for priorities. *)
 type 'a scheduler
 
+(** Panic exception is raised when duppy's event loops has
+  * crashed. This exception should be considered a MAJOR
+  * FAILURE. All current non-ready tasks registered for the 
+  * calling scheduler are dropped. You may restart Duppy's queues after it is raised but
+  * it should only be used to terminate the process diligently!! *)
+exception Panic of exn
+
 (** Initiate a new scheduler 
   * @param compare the comparison function used to sort tasks according to priorities. 
   * Works as in [List.sort] *)
@@ -65,6 +74,9 @@ val create : ?compare:('a -> 'a -> int) -> unit -> 'a scheduler
  * starts a queue, on the scheduler [s] only processing priorities [p]
  * for which [priorities p] returns [true].
  *
+ * Raises [Panic exception] if [exception] was raised while executing
+ * the main event loop.
+ *
  * Several queues can be run concurrently against [s]. 
  * @param log Logging function. Default: [Printf.printf "queue %s: %s\n" name]
  * @param priorities Predicate specifying which priority to process. Default: [fun _ -> _ -> true]*)
@@ -95,7 +107,10 @@ sig
     handler  : 'b list -> ('a,'b) task list
   }
 
-  (** Type for possible events. *)
+  (** Type for possible events. 
+    *
+    * Please not that currently, under win32, all socket used in ocaml-duppy 
+    * are expected to be in blocking mode only! *)
   type event = [
     | `Delay of float
     | `Write of Unix.file_descr
@@ -301,29 +316,43 @@ sig
     * the calling computation and not the calling thread. *)
   module Mutex : 
   sig
-    (** Type for a mutex. *)
-    type mutex 
+    (** Information used to initialize a Mutex module. 
+      * [priority] and [scheduler] are used to initialize a task
+      * which treat mutexes as well as conditions from the below
+      * [Condition] module. *)
+    module type Mutex_control =
+    sig
+      type priority
+      val scheduler : priority scheduler
+      val priority : priority
+    end
 
-    (** [create ~priority s] creates a mutex. Implementation-wise,
-      * a duppy task is created that will be used to select a 
-      * waiting computation, lock the mutex on it and resume it.
-      * Thus, [priority] and [s] represents, resp., the priority 
-      * and scheduler used when running calling process' computation. *)
-    val create : priority:'a -> 'a scheduler -> mutex
+    module type Mutex_t =
+    sig
+      (** Type for a mutex. *)
+      type mutex
 
-    (** A computation that locks a mutex
-      * and returns [unit] afterwards. Computation
-      * will be blocked until the mutex is sucessfuly locked. *)
-    val lock : mutex -> (unit,'a) t
+      module Control : Mutex_control
 
-    (** A computation that tries to lock a mutex.
-      * Returns immediatly [true] if the mutex was sucesfully locked
-      * or [false] otherwise. *)
-    val try_lock : mutex -> (bool,'a) t
+      (** [create ()] creates a mutex. *)
+      val create : unit -> mutex
 
-    (** A computation that unlocks a mutex. 
-      * Should return immediatly. *)
-    val unlock : mutex -> (unit,'a) t
+      (** A computation that locks a mutex
+        * and returns [unit] afterwards. Computation
+        * will be blocked until the mutex is sucessfuly locked. *)
+      val lock : mutex -> (unit,'a) t
+
+      (** A computation that tries to lock a mutex.
+        * Returns immediatly [true] if the mutex was sucesfully locked
+        * or [false] otherwise. *)
+      val try_lock : mutex -> (bool,'a) t
+
+      (** A computation that unlocks a mutex.
+        * Should return immediatly. *)
+      val unlock : mutex -> (unit,'a) t
+    end
+
+    module Factory(Control : Mutex_control) : Mutex_t 
   end
 
   (** This module implements monadic
@@ -335,34 +364,37 @@ sig
     * been called. *)
   module Condition : 
   sig
-    (** Type of a condition, used in [wait] and [broadcast] *)
-    type condition
+    module Factory(Mutex : Mutex.Mutex_t) :
+    sig
+      (** Type of a condition, used in [wait] and [broadcast] *)
+      type condition
 
-    (** Create a condition. Implementation-wise,
-      * a duppy task is created that will be used to select a
-      * waiting computation, and resume it.
-      * Thus, [priority] and [s] represents, resp., the priority
-      * and scheduler used when running calling process' computation. *)
-    val create : priority:'a -> 'a scheduler -> condition
+      (** Create a condition. Implementation-wise,
+        * a duppy task is created that will be used to select a
+        * waiting computation, and resume it.
+        * Thus, [priority] and [s] represents, resp., the priority
+        * and scheduler used when running calling process' computation. *)
+      val create : unit -> condition
 
-    (** [wait h m] is a computation that:
-      * {ul
-      * {- Unlock mutex [m]}
-      * {- Wait until [Condition.signal c] or [Condition.broadcast c] 
-           has been called}
-      * {- Locks mutex [m]}
-      * {- Returns [unit]}} *)
-    val wait : condition -> Mutex.mutex -> (unit,'b) t
-
-    (** [broadcast c] is a computation that 
-      * resumes all computations waiting on [c]. It should
-      * return immediatly. *)
-    val broadcast : condition -> (unit,'a) t
-
-    (** [signal c] is a computation that resumes one 
-      * computation waiting on [c]. It should return
-      * immediatly. *)
-    val signal : condition -> (unit,'a) t
+      (** [wait h m] is a computation that:
+        * {ul
+        * {- Unlock mutex [m]}
+        * {- Wait until [Condition.signal c] or [Condition.broadcast c] 
+             has been called}
+        * {- Locks mutex [m]}
+        * {- Returns [unit]}} *)
+      val wait : condition -> Mutex.mutex -> (unit,'a) t
+  
+      (** [broadcast c] is a computation that 
+        * resumes all computations waiting on [c]. It should
+        * return immediately. *)
+      val broadcast : condition -> (unit,'a) t
+  
+      (** [signal c] is a computation that resumes one 
+        * computation waiting on [c]. It should return
+        * immediately. *)
+      val signal : condition -> (unit,'a) t
+    end
   end
 
   (** This module implements monadic computations
